/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005-2008. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("com.ibm.team.repository.web.ui.internal.admin.server.LicenseEditor");

dojo.require("dojo.date");
dojo.require("dojo.date.locale");
dojo.require("dojo.date.stamp");
dojo.require("dojo.number");
dojo.require("dijit._Widget");
dojo.require("dijit._Templated");
dojo.require("dojox.data.dom");

dojo.require("net.jazz.ajax.ui.internal.HoverPopup");
dojo.require("net.jazz.ajax.ui.PlatformUI");

dojo.require("com.ibm.team.repository.web.client.internal.licensing");
dojo.require("com.ibm.team.repository.web.transport.ServiceResponseHandler");
dojo.require("com.ibm.team.repository.web.ui.internal.admin.server.FileUploader");
dojo.require("com.ibm.team.repository.web.ui.internal.Util");

//
// Aliases
//

var FileUploader = com.ibm.team.repository.web.ui.internal.admin.server.FileUploader;
var LIC_P = com.ibm.team.repository.web.client.internal.licensing;
var formatCALTypeDescription = LIC_P.formatCALTypeDescription;
var getLicenseInfo = LIC_P.getLicenseInfo;
var getExpirationInfo = LIC_P.getExpirationInfo;
var sortCALs = LIC_P.sortCALs;
var HoverPopup = net.jazz.ajax.ui.internal.HoverPopup;
var postLicenseInfo = LIC_P.postLicenseInfo;
var removeChildren = dojox.data.dom.removeChildren;
var SRH = com.ibm.team.repository.web.transport.ServiceResponseHandler;
var UPLOAD_URI = LIC_P.FILE_UPLOAD_URI;

//
// Messages
//

var MSG_REFRESH_DIRTY_FORM = "Refreshing will result in losing your changes. Continue?";

//
// Widget definition
//

dojo.declare("com.ibm.team.repository.web.ui.internal.admin.server.LicenseEditor", [dijit._Widget, dijit._Templated], {
	
	//
	// Class properties
	//
	
	templatePath: dojo.moduleUrl("com.ibm.team.repository.web", "ui/internal/admin/server/templates/LicenseEditor.html"),
	
	//
	// Widget lifecycle methods
	//
	
	postCreate: function() {
		this.loadForms();
	}, 
	
	/**
	 * Fetches the most recent saved license information from the server and
	 * resets the editor with this information (long-op).
	 */
	reload: function() {
		if(this._isDirty && !confirm(MSG_REFRESH_DIRTY_FORM)){
			return;
		}
		getWorkbench().showMessage("Loading license information");
		var self = this;
		var o = {
			s: function(licenseInfo) {
				getWorkbench().hideMessageArea();
				self.reset(licenseInfo);
			},
			f: function(err) {self.cbRemoteError(err);}
		}
		getLicenseInfo(new SRH(o, "s", "f"));
	},
	
	rebind: function() {
		this.refreshServerTable();
		this.refreshCALTable();
	},
	
	/**
	 * Resets both the object and UI state of this editor, based on the
	 * <code>licenseInfo</code> parameter (short-op).
	 */
	reset: function(licenseInfo) {
		// reset object state
		this.initialLicenseInfo = this.licenseInfo = licenseInfo;
		this._isDirty = false;
		this._uploadedCALFiles = {};
		this._uploadedServerFile = null;
		
		// reset UI state
		this.hideMessageArea();
		this.dirtyMarker.innerHTML = "";
		this.rebind();
		this.saveButton.disabled = true;
	},
	
	/**
	 * Saves the current state of the editor (long-op).
	 */
	save: function() {
		this.saveButton.disabled = true;
		getWorkbench().showMessage("Saving license information...");
		var args = dojo.mixin(this.getEditorData(), {commit: true});
		postLicenseInfo(new SRH(this, "cbSave", "cbSaveError"), args);
	},
	
	/**
	 * Success callback from 'Save' action; resets editor based on newly
	 * persisted state.
	 */
	cbSave: function(licenseInfo) {
		this.reset(licenseInfo);
		function f() {
			getWorkbench().hideMessageArea();
		}
		dojo.publish(com.ibm.team.repository.web.internal.events.licenseChange, [licenseInfo]);
		this.showEventOutcomeMessage("ok", "Save complete", f, 1500);
	},
	
	cbSaveError: function(errObj) {
		this.saveButton.disabled = false;
		this.cbRemoteError(errObj, "A problem occurred and the license configuration was not saved");
	},
	
	/**
	 * Called by Admin Web UI when this cached widget is displayed; update
	 * the form's backing data to make sure we pick up any changes to user
	 * assignments.
	 */
	show: function() {
		if(this._isDirty) {
			this.fetchToBeState();
		} else {
			this.reload();
		}
	},
	
	//
	// UI data binding and querying
	//
	
	loadForms: function() {
		var formName1 = "newCALUploadForm";
		this[formName1] = this.createForm("CAL", "Client Activation License Key", formName1);
		
		var formName2 = "serverUploadForm";
		this[formName2]	= this.createForm("server", "Server Activation Key", formName2);
	},
	
	createForm: function(typeCode, typeName, formName) {
		var self = this;
		var args = {
			formElements: [{name: "licenseType", value: typeCode}],
			formLabel: "Upload Key: ",
			onBeginUpload: function() { getWorkbench().showMessage("Uploading file"); },
			onUpload: function(o) { self.cbUpload(typeCode, typeName, formName, o.fileName); },
			onUploadFailure: function(errObj){ self.cbUploadFailure(typeName, errObj); },
			postURI: UPLOAD_URI
		}
		return new FileUploader(args, this[formName]);
	}, 
	
	/**
	 * Returns the current data associated with the editor needed to call
	 * <code>postLicenseInfo()</code>.
	 */
	getEditorData: function() {
		var calFiles = [];
		for (var fileName in this._uploadedCALFiles) {
			calFiles.push(fileName);
		}
		var args = {
			serverFile: this._uploadedServerFile,
			calFiles: calFiles,
			maxCounts: []
		};
		
		// if this is not a trial, possibly update the purchased counts
		var inputs = this.CALInfoTable.getElementsByTagName("input");
		dojo.forEach(inputs, function(input){
			if(!input.disabled) {
				args.maxCounts.push({id: input.__jazzCALId, max: toInt(input.value)});
			}
		});

		return args;
	},
	
	/**
	 * Refreshes the Client Access License table based on the current value of
	 * <code>this.licenseData.types</code>.
	 */
	refreshCALTable: function() {
		// clear current CAL table values
		var tbody = this.CALInfoTable;
		var self = this;
		var inputs = this.CALInfoTable.getElementsByTagName("input");
		dojo.forEach(inputs, function(input) {
			if(input.__jazzChgEvt){
				self.disconnect(input.__jazzChgEvt);
			}
		});
		removeChildren(tbody);
		
		// normalize data and create table columns
		var types = this.licenseInfo.types || [];
		types = sortCALs(types);
		var isTrial = this.licenseInfo.serverLicense.expirationTime ? true : false;
		removeChildren(this.purchasedOrTrialHeader);
		this.purchasedOrTrialHeader.appendChild(document.createTextNode(isTrial ? "Trial" : "Purchased"));
		dojo.forEach(types, function(type){
			var row = d.ce("tr");
			tbody.appendChild(row);
			// deal with disabled CALs
			var isDisabled = type.disabledReason != null;
			if(isDisabled) {
				row.className = "disabledCAL";
			}
			// deal wth unlimited CALs
			var isUnlimited = type.max === -1;
			if(isUnlimited) {
				var limit = "Unlimited";
				var included = "N/A";
			} else {
				var limit = type.max;
				var included = type.free;
			}
			
			// set displayable counts depending on whether the product is a
			// permanent edition or a trial
			if (isUnlimited) {
				var total = "Unlimited";
				var available = "Unlimited";
				var purchasedOrTrialVal = "N/A";
			} else if (isTrial) {
				var total = type.max;
				var available = type.max - type.used;
				var purchasedOrTrialVal = type.max - type.free;
			} else {
				var total = Math.min(type.max, type.total);
				var available = total - type.used;
				var purchasedOrTrialVal = type.total - type.free;
			}
			// create the columns for the given CAL
			var nameLink = d.ce("a");
			nameLink.href = "#";
			nameLink.onclick = function(){
				return false;
			};
			nameLink.innerHTML = type.name;
			var description = formatCALTypeDescription(type);
			var tooltip = new HoverPopup({
				connectNode: [ nameLink ],
				label: description,
				showDelay: 300,
				orientation: "R"
			}, d.ce("span"));
			var nodes = [
				type.productName,
				type.editionName + " (" + type.variantName + ")",
				nameLink,
				limit,
				included,
				// purchased or trial count
				purchasedOrTrialCol(self, type, isTrial, isUnlimited, purchasedOrTrialVal),
				total,
				getUsedLink(type),
				available,
				isDisabled ? "Deactivated (" + type.disabledReason + ")" : "Activated"
			];
			dojo.forEach(nodes, function(node) {
				col(row, node);
			});
		});
	},
	
	/**
	 * Refreshes the Server License information table based on the current value
	 * of <code>this.licenseData.serverLicense</code>.
	 */
	refreshServerTable: function(serverLicense) {
		var serverLicense = this.licenseInfo.serverLicense;
		removeChildren(this.serverInfoTable);
		var tbody = this.serverInfoTable;
		var isTrial = serverLicense.expirationTime ? true : false;
		if(isTrial) {
			var expirationInfo = getExpirationInfo(serverLicense.expirationTime);
		}
		var isDisabled = serverLicense.disabledReason || (isTrial && expirationInfo.isExpired);
		function row(propRow, className) {
			var row = d.ce("tr");
			row.className = isDisabled ? "disabledServer" : "";
			tbody.appendChild(row);
			var td = d.ce("td");
			row.appendChild(td);
			td.innerHTML = propRow.n;
			td.className = "propColumn";
			var td = d.ce("td");
			row.appendChild(td);
			td.innerHTML = propRow.v;
			if(className) {
				td.className = className;
			}
		}
		var maxUsers = serverLicense.maxContributors !== -1 ? serverLicense.maxContributors : "Unlimited";
		var editionLabel = serverLicense.editionName ? serverLicense.editionName : "";
		editionLabel += serverLicense.variantName ? " (" + serverLicense.variantName + ")" : ""; 
		var rows = [
			{n: "Product", v: serverLicense.productName},
			{n: "Edition", v: editionLabel},
			{n: "Description", v: serverLicense.description},
			{n: "Maximum Users", v: maxUsers},
			{n: "Users with Counted Licenses", v: serverLicense.usedContributors}
		];
		dojo.forEach(rows, function(propRow){
			row(propRow);
		});

		var className = null;
		if(serverLicense.disabledReason) {
			var statusString = serverLicense.disabledReason;
		} else {
			var statusString = !isTrial ? "Permanent" : expirationInfo.isExpired ? "Expired Trial" : "Trial";
		}
		row({n: "Status", v: statusString}, isDisabled ? "expired" : null);
		
		if(isTrial) {
			if(expirationInfo.isExpired) {
				var rowObj = {n: "Expired On", v: expirationInfo.expirationDateString + " (" + expirationInfo.daysSinceExpirationString + ")"};
			} else {
				var rowObj = {n: "Expires On", v: expirationInfo.expirationDateString + " (" + expirationInfo.daysRemainingString + ")"};
			}
			row(rowObj, isDisabled ? "expired" : null);
		}
	},
	
	//
	// UI state manipulation methods
	//
	
	markDirty: function() {
		this.dirtyMarker.innerHTML = "*";
		this._isDirty = true;
	},
	
	//
	// Utility methods
	//
	
	/**
	 * General purpose error callback for remote methods.
	 */
	cbRemoteError: function(errorObj, msg) {
		if(!msg) {
			msg = "There was a problem communicating with the server";
		}
		if(errorObj.status === 0) {
			msg += " (Server not reachable)";
		} else {
			msg += " (" + errorObj.message + ")";
		}
		this.showEventOutcomeMessage("error", msg);
	},
	
	//
	// UI Status and State reporting
	//
	
	updateSaveButton: function() {
		this.saveButton.disabled = !this._isDirty || this.status.errors.length > 0;
	},
	
	showEventOutcomeMessage: function(type, message, cb, duration) {
		if(cb){
			setTimeout(cb, duration || 3000);
		} else if (duration) {
			setTimeout(function(){getWorkbench().hideMessageArea();}, duration);
		}
		getWorkbench().showMessage(message);
	},
	
	updateStatusMessage: function() {
		var hasErrors = this.status.errors.length > 0;
		var hasWarnings = this.status.warnings.length > 0;
		if(hasErrors || hasWarnings) {
			if (hasErrors) {
				this.showMessageArea("error", this.status.errors);
			} else {
				this.showMessageArea("warning", this.status.warnings);
			}
		} else if (this._isDirty) {
			this.showMessageArea("info", ["The license configuration has pending changes. Press 'Save' to persist the changes or press the refresh icon to discard the changes."]);
		} else {
			this.hideMessageArea();
		}
	},
	
	showMessageArea: function(type, messageNodes) {
		removeChildren(this.messageArea);
		if(!dojo.isArrayLike(messageNodes)) {
			messageNodes = [messageNodes];
		}
		this.messageArea.className = "messageArea " + type;
		var self = this;
		dojo.forEach(messageNodes, function(node){
			if(dojo.isString(node)) {
				node = document.createTextNode(node);
			}
			self.messageArea.appendChild(node);
		});
		this.messageArea.style.visibility = "visible";
	},
	
	hideMessageArea: function(cb) {
		this.messageArea.style.visibility = "hidden";
		this.messageArea.innerHTML = "Messages go here!";
	},
	
	//
	// General UI code
	//
	
	//
	// Validation logic and UI - validation can be triggered in two ways:
	// 1) User uploads a file (cbUpload calls fetchToBeState)
	// 2) User enters a new purchased count for a CAL (onPurchasedCountKeyUp)
	//
	
	fetchToBeState: function() {
		getWorkbench().showMessage("Validating license information");
		var args = dojo.mixin(this.getEditorData(), {commit: false});
		postLicenseInfo(new SRH(this, "cbFetchToBeState", "cbRemoteError"), args);
	},
	
	cbFetchToBeState: function(licenseInfo) {
		getWorkbench().hideMessageArea();
		this.licenseInfo = licenseInfo;
		this.rebind();
		this.validate();
	},
	
	onPurchasedCountKeyUp: function(e) {
		var input = e.target;
		if(!input || input.__jazzOrigValue === undefined) {
			return; // early return
		}
		
		if(input.value !== input.__jazzOrigValue) {
			this.markDirty();
		}
		// run a validation after the user stops typing for half a second
		if(this._calKeyStrokeTimout){
			clearTimeout(this._calKeyStrokeTimout);
		}
		this._calKeyStrokeTimout = setTimeout(dojo.hitch(this, this.validate), 500);
	},
	
	/**
	 * Validate does two things:
	 * 
	 * 1) Compares the state of the CALs in the dirty editor vs. the CALs'
	 *    original state. The purpose of this validation is to generate warnings
	 *    if the user is about to disable users that were previously enabled.
	 * 2) Ensures that editable CAL 'purchased' counts are >= 0 and less than
	 *    the maximum purchasable count for the CAL.
	 */
	validate: function() {
		
		// currently, only CAL purchase counts can have errors and only CAL
		// state changes can produce warnings
		this.status = {
			errors: this.validateCALPurchaseCounts(),
			warnings: this.validateCALStates()
		};
		this.updateSaveButton();
		this.updateStatusMessage();
	},
	
	validateCALStates: function() {
		var warnings = [];
		var origGood = {};
		var origTypes = this.initialLicenseInfo.types;
		dojo.forEach(origTypes, function(type){
			if(!type.disabledReason) {
				origGood[type.id] = true;
			}
		});
		var dirtyTypes = this.licenseInfo.types;
		dojo.forEach(dirtyTypes, function(type){
			if(type.disabledReason && origGood[type.id]) {
				warnings.push("Client Access Type '" + type.name + "' will be deactivated. See 'Client Access Licenses Activation Keys' section for more information.");
			}
		});
		return warnings;
	},
	
	validateCALPurchaseCounts: function() {
		var errors = [];
		var inputs = this.CALInfoTable.getElementsByTagName("input");
		var isValid = true;
		for(var i=0; i<inputs.length; i++) {
			if(inputs[i].disabled) {
				continue;
			}
			var result = validateCALInput(inputs[i]);
			if(!result.valid) {
				errors = ["Please fix the problems with the Client Access License Activation Keys purchase counts below"];
			}
		}
		return errors;
	},
	
	//
	// Upload-related methods
	//
	
	cbUpload: function(typeCode, typeName, widgetPropName, fileName) {
		this.markDirty();
		if(typeCode === "CAL") {
			this._uploadedCALFiles[fileName] = true;
		} else {
			this._uploadedServerFile = fileName;
		}
		this.showEventOutcomeMessage("ok", "Successfully uploaded " + typeName, dojo.hitch(this, this.fetchToBeState), 1000);
	},
	
	cbUploadFailure: function(typeName, errObj) {
		this.showEventOutcomeMessage("error", "Error uploading " + typeName + ": " + errObj.errorMessage);
	}
});

//
// Utility functions
//

var d = document;
d.ce = document.createElement;

function clearMessageTimeout(ctxObj, timeoutProp) {
	var timeoutObj = ctxObj[timeoutProp];
	clearTimeout(timeoutObj);
	ctxObj[timeoutProp] = null;
}

function col(row, node) {
	if(node === null || node === undefined) {
		node = "-";
	}
	var td = d.ce("td");
	var isNumber = typeof node == 'number' || node instanceof Number;
	var isNumberCol = isNumber || node === "N/A" || node === "Unlimited";
	if(isNumberCol) {
		td.className = "num";
	}
	if(isNumber) {
		node = dojo.number.format("" + node);
	}
	var isArray = !dojo.isString(node) && dojo.isArrayLike(node);
	if(node.tagName === "A" && node.className === "num") {
		td.className = "num";
	} else if (isArray) {
		td.className = "num inputCol";
	}
	row.appendChild(td);
	var nodeArr = isArray ? node : [node];
	dojo.forEach(nodeArr, function(node){
		if(dojo.isString(node)) {
			node = document.createTextNode(node);
		}
		td.appendChild(node);
	});
}

function purchasedOrTrialCol(widget, type, isTrial, isUnlimited, purchasedOrTrial) {
	var input = d.ce("input");
	var msgDiv = d.ce("div");
	input.className = "num";
	input.value = input.__jazzOrigValue = "" + purchasedOrTrial;
	input.__jazzCALId = type.id;
	input.__jazzMaxPurchasable = type.max - type.free >= 0 ? type.max - type.free : 0;
	input.__jazzMsgArea = msgDiv;
	if(!isTrial && !isUnlimited && (type.max - type.free > 0)) {
		input.__jazzChgEvt = widget.connect(input, "onchange", widget.validate);
	} else {
		input.disabled = true;
	}
	return [input, msgDiv];
}

function getCountString(count) {
	return count === -1 ? "Unlimited" : (dojo.number.format(count)); 
}

function getUsedLink(calType) {
	if(calType.used === 0) {
		return 0;
	} else {
		var a = document.createElement("a");
		a.href = "#action=com.ibm.team.repository.manageUserCALs&licenseType=" + calType.id;
		a.innerHTML = calType.used;
		a.className = "num";
		return a;
	}
}

function getWorkbench() {
	return net.jazz.ajax.ui.PlatformUI.getWorkbench();
}

var NUM_REGEX = /[0-9]+/g;

function toInt(text) {
	var match = text.match(NUM_REGEX);
	return match && parseInt(text, 10);
}

function validateCALInput(input) {
	var numVal = toInt(input.value);
	if(numVal !== null && numVal <= input.__jazzMaxPurchasable && numVal >= 0) {
		isValid = true;
		input.__jazzMsgArea.innerHTML = "";
	} else {
		var isValid = false;
		input.__jazzMsgArea.innerHTML = "The purchase count for this key must be a number between 0 and " + input.__jazzMaxPurchasable;
	}
	var node = input;
	while(node.tagName !== "TR") {
		if(node.tagName === "TD") {
			node.className = isValid ? "num inputCol" : "inputCol";
		}
		node = node.parentNode;
	}
	if(isValid) {
		dojo.removeClass(node, "hasError");
	} else {
		dojo.addClass(node, "hasError");
	}
	return {
		changed: input.value !== input.__jazzOrigValue,
		valid: isValid
	};
}

})();